﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2011 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.PermutationEncoding;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Problems.QuadraticAssignment {
  [Item("QAPSwapMoveEvaluator", "Evaluated a swap move on a QAP solution.")]
  [StorableClass]
  public class QAPSwapMoveEvaluator : QAPMoveEvaluator, IPermutationSwapMoveOperator {
    public ILookupParameter<SwapMove> SwapMoveParameter {
      get { return (ILookupParameter<SwapMove>)Parameters["SwapMove"]; }
    }

    [StorableConstructor]
    protected QAPSwapMoveEvaluator(bool deserializing) : base(deserializing) { }
    protected QAPSwapMoveEvaluator(QAPSwapMoveEvaluator original, Cloner cloner)
      : base(original, cloner) {
    }
    public QAPSwapMoveEvaluator() {
      Parameters.Add(new LookupParameter<SwapMove>("SwapMove", "The move to evaluate."));
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new QAPSwapMoveEvaluator(this, cloner);
    }

    public static double Apply(Permutation assignment, SwapMove move, DoubleMatrix weights, DoubleMatrix distances) {
      if (move.Index1 == move.Index2) return 0;
      double moveQuality = 0;
      for (int j = 0; j < assignment.Length; j++) {
        moveQuality -= weights[move.Index1, j] * distances[assignment[move.Index1], assignment[j]];
        moveQuality -= weights[j, move.Index1] * distances[assignment[j], assignment[move.Index1]];
        moveQuality -= weights[move.Index2, j] * distances[assignment[move.Index2], assignment[j]];
        moveQuality -= weights[j, move.Index2] * distances[assignment[j], assignment[move.Index2]];

        int facility1 = j == move.Index1 ? assignment[move.Index2] : assignment[j];
        int facility2 = j == move.Index2 ? assignment[move.Index1] : assignment[j];

        moveQuality += weights[move.Index1, j] * distances[assignment[move.Index2], facility2];
        moveQuality += weights[j, move.Index1] * distances[facility2, assignment[move.Index2]];
        moveQuality += weights[move.Index2, j] * distances[assignment[move.Index1], facility1];
        moveQuality += weights[j, move.Index2] * distances[facility1, assignment[move.Index1]];
      }
      return moveQuality;
    }

    public override IOperation Apply() {
      SwapMove move = SwapMoveParameter.ActualValue;
      if (move == null) throw new InvalidOperationException("Swap move is not found.");
      Permutation assignment = PermutationParameter.ActualValue;
      DoubleMatrix distances = DistanceMatrixParameter.ActualValue;
      DoubleMatrix weights = WeightsParameter.ActualValue;

      double moveQuality = QualityParameter.ActualValue.Value;
      moveQuality += Apply(assignment, move, weights, distances);
      MoveQualityParameter.ActualValue = new DoubleValue(moveQuality);
      return base.Apply();
    }
  }
}
